# -*- coding: utf-8 -*-
from django.utils.translation import ugettext_lazy as _

from model_utils import Choices


QUALIFIERS = Choices(
    (1, 'BILATERAL', _('BILATERAL')),
    (2, 'CHRONIC', _('CHRONIC')),
    (3, 'DISTAL', _('DISTAL')),
    (4, 'EPISODIC', _('EPISODIC')),
    (5, 'GENERALIZED', _('GENERALIZED')),
    (6, 'MILD', _('MILD')),
    (7, 'MODERATE', _('MODERATE')),
    (8, 'NONPROGRESSIVE', _('NONPROGRESSIVE')),
    (9, 'PROFOUND', _('PROFOUND')),
    (10, 'PROGRESSIVE', _('PROGRESSIVE')),
    (11, 'PROXIMAL', _('PROXIMAL')),
    (12, 'RECURRENT', _('RECURRENT')),
    (13, 'REFRACTORY', _('REFRACTORY')),
    (14, 'SEVERE', _('SEVERE')),
)

CROSS_REFERENCES = Choices(
    (1, 'UMLS', _('UMLS')),
    (2, 'MSH', _('MSH')),
    (3, 'SNOMEDCT_US', _('SNOMEDCT_US')),
    (4, 'MEDDRA', _('MEDDRA')),
    (5, 'ICD-10', _('ICD-10')),
    (6, 'EPCC', _('EPCC')),
    (7, 'ICD-O', _('ICD-O')),
    (8, 'DOID', _('DOID')),
    (9, 'MP', _('MP')),
    (10, 'MPATH', _('MPATH')),
    (11, 'PMID', _('PMID')),
    (12, 'NCIT', _('NCIT')),
    (13, 'DOI', _('DOI')),
    (14, 'ORPHA', _('ORPHA')),
    (15, 'ICD-9', _('ICD-9')),
    (16, 'RGD', _('RGD')),
)

DATABASES = Choices(
    (1, 'OMIM', _('OMIM')),
    (2, 'ORPHA', _('ORPHA')),
    (3, 'DECIPHER', _('DECIPHER')),
)

EVIDENCE = Choices(
    (1, 'ICE', _('ICE')),
    (2, 'IEA', _('IEA')),
    (3, 'PCS', _('PCS')),
    (4, 'TAS', _('TAS')),
)

SYNONYM_SCOPES = Choices(
    (1, 'EXACT', _('EXACT')),
    (2, 'BROAD', _('BROAD')),
    (3, 'NARROW', _('NARROW')),
    (4, 'RELATED', _('RELATED')),
)



# -*- coding: utf-8 -*-
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import viewsets
from rest_framework.filters import SearchFilter

from genome.models import Gene
from genome.serializers import GeneSerializer

from . import filters, models, serializers


class TermViewSet(viewsets.ReadOnlyModelViewSet):
    """ViewSet for viewing HPO Terms."""

    queryset = models.Term.objects.fast()
    serializer_class = serializers.TermSerializer
    filter_backends = (DjangoFilterBackend, SearchFilter)
    filter_class = filters.TermFilter
    search_fields = (
        'label',
        'description',
        'synonyms__description',
    )


class CrossReferenceViewSet(viewsets.ReadOnlyModelViewSet):
    """ViewSet for viewing HPO Term Cross References."""

    queryset = models.CrossReference.objects.fast()
    serializer_class = serializers.CrossReferenceSerializer
    filter_backends = (DjangoFilterBackend, SearchFilter)
    filter_class = filters.CrossReferenceFilter
    search_fields = ('source_value', )


class DiseaseViewSet(viewsets.ReadOnlyModelViewSet):
    """ViewSet for viewing Diseases."""

    queryset = models.Disease.objects.fast()
    serializer_class = serializers.DiseaseSerializer
    filter_backends = (DjangoFilterBackend, SearchFilter)
    filter_class = filters.DiseaseFilter
    search_fields = (
        'description',
        'hpo_terms__label',
        'hpo_terms__description',
        'hpo_terms__synonyms__description',
    )


class DiseaseGeneViewSet(viewsets.ReadOnlyModelViewSet):
    """A simple ViewSet for viewing Disease Genes."""

    queryset = Gene.objects \
        .prefetch_related('diseases__hpo_terms') \
        .filter(diseases__id__isnull=False) \
        .distinct()
    serializer_class = GeneSerializer
    filter_backends = (DjangoFilterBackend, SearchFilter)
    filter_class = filters.DiseaseGeneFilter
    search_fields = (
        'diseases__description',
        'diseases__hpo_terms__label',
        'diseases__hpo_terms__description',
        'diseases__hpo_terms__synonyms__description',
    )



# -*- coding: utf-8 -*-
# Generated by Django 1.11.8 on 2018-01-09 14:17
from __future__ import unicode_literals

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields
import re


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('genome', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='CrossReference',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('source', models.PositiveSmallIntegerField(choices=[(1, 'UMLS'), (2, 'MSH'), (3, 'SNOMEDCT_US'), (4, 'MEDDRA'), (5, 'ICD-10'), (6, 'EPCC'), (7, 'ICD-O'), (8, 'DOID'), (9, 'MP'), (10, 'MPATH'), (11, 'PMID'), (12, 'NCIT'), (13, 'DOI'), (14, 'ORPHA'), (15, 'ICD-9'), (16, 'RGD')])),
                ('source_value', models.CharField(max_length=150)),
            ],
            options={
                'verbose_name': 'XRef',
                'verbose_name_plural': 'XRefs',
            },
        ),
        migrations.CreateModel(
            name='Disease',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('database', models.PositiveSmallIntegerField(choices=[(1, 'OMIM'), (2, 'ORPHA'), (3, 'DECIPHER')])),
                ('identifier', models.IntegerField()),
                ('description', models.TextField(blank=True)),
                ('qualifier', models.PositiveSmallIntegerField(blank=True, choices=[(1, 'BILATERAL'), (2, 'CHRONIC'), (3, 'DISTAL'), (4, 'EPISODIC'), (5, 'GENERALIZED'), (6, 'MILD'), (7, 'MODERATE'), (8, 'NONPROGRESSIVE'), (9, 'PROFOUND'), (10, 'PROGRESSIVE'), (11, 'PROXIMAL'), (12, 'RECURRENT'), (13, 'REFRACTORY'), (14, 'SEVERE')], null=True)),
                ('evidence_code', models.PositiveSmallIntegerField(blank=True, choices=[(1, 'ICE'), (2, 'IEA'), (3, 'PCS'), (4, 'TAS')], null=True)),
                ('create_date', models.DateField()),
                ('assigned_by', models.CharField(max_length=255)),
            ],
            options={
                'verbose_name': 'Disease',
                'verbose_name_plural': 'Diseases',
            },
        ),
        migrations.CreateModel(
            name='Synonym',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.TextField(blank=True)),
                ('scope', models.PositiveSmallIntegerField(blank=True, choices=[(1, 'EXACT'), (2, 'BROAD'), (3, 'NARROW'), (4, 'RELATED')], null=True)),
            ],
            options={
                'verbose_name': 'Synonym',
                'verbose_name_plural': 'Synonyms',
            },
        ),
        migrations.CreateModel(
            name='Term',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('identifier', models.PositiveIntegerField(db_index=True)),
                ('label', models.CharField(blank=True, max_length=255)),
                ('description', models.TextField(blank=True)),
                ('created_by', models.CharField(blank=True, max_length=50)),
                ('created', models.CharField(blank=True, max_length=25)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('alternate_ids', models.CharField(blank=True, max_length=500, validators=[django.core.validators.RegexValidator(re.compile('^\\d+(?:\\,\\d+)*\\Z', 32), code='invalid', message='Enter only digits separated by commas.')])),
                ('can_be', models.ManyToManyField(blank=True, related_name='_term_can_be_+', to='hpo_terms.Term')),
                ('is_a', models.ManyToManyField(blank=True, related_name='_term_is_a_+', to='hpo_terms.Term')),
            ],
            options={
                'verbose_name': 'Term',
                'verbose_name_plural': 'Terms',
            },
        ),
        migrations.CreateModel(
            name='Version',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('label', models.CharField(max_length=100)),
            ],
            options={
                'verbose_name': 'Version',
                'verbose_name_plural': 'Versions',
            },
        ),
        migrations.AddField(
            model_name='term',
            name='version',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='hpo_terms.Version'),
        ),
        migrations.AddField(
            model_name='synonym',
            name='hpo_term',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='synonyms', to='hpo_terms.Term'),
        ),
        migrations.AddField(
            model_name='disease',
            name='age_of_onset',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='disease_age_of_onsets', to='hpo_terms.Term'),
        ),
        migrations.AddField(
            model_name='disease',
            name='frequency',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='disease_frequencies', to='hpo_terms.Term'),
        ),
        migrations.AddField(
            model_name='disease',
            name='genes',
            field=models.ManyToManyField(blank=True, related_name='diseases', to='genome.Gene'),
        ),
        migrations.AddField(
            model_name='disease',
            name='hpo_terms',
            field=models.ManyToManyField(blank=True, related_name='diseases', to='hpo_terms.Term'),
        ),
        migrations.AddField(
            model_name='disease',
            name='mode_of_inheritance',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='disease_mois', to='hpo_terms.Term'),
        ),
        migrations.AddField(
            model_name='disease',
            name='version',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='hpo_terms.Version'),
        ),
        migrations.AddField(
            model_name='crossreference',
            name='hpo_term',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='xrefs', to='hpo_terms.Term'),
        ),
        migrations.AlterIndexTogether(
            name='term',
            index_together=set([('version', 'identifier')]),
        ),
        migrations.AlterUniqueTogether(
            name='disease',
            unique_together=set([('version', 'database', 'identifier')]),
        ),
        migrations.AlterIndexTogether(
            name='disease',
            index_together=set([('database', 'identifier')]),
        ),
    ]



# -*- coding: utf-8 -*-
def normalize_hpo_id(hpo_term):
    """Reformat HPO Term to database Id

    Arguments:
        hpo_term (str): HPO Term

    Returns:
        int: HPO database Id
    """
    if hpo_term.startswith('HP:'):
        hpo_id = hpo_term.lstrip('HP:')
    else:
        hpo_id = hpo_term

    return int(hpo_id.lstrip('0'))



# -*- coding: utf-8 -*-
import datetime
import logging

from pronto import Ontology

from django.core.exceptions import ObjectDoesNotExist
from django.core.management import BaseCommand
from django.db.models import Q

from genome.models import Gene
from genomix.utils import retrieve_data

from hpo_terms import app_settings, choices, models, utils


logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = 'Sync HPO from online resource'

    def handle(self, *args, **options):
        # Download external resources
        logger.info('Downloading HPO resources...')

        hpo = Ontology(app_settings.HPO_URL, timeout=10)
        hpo_diseases = retrieve_data(app_settings.HPO_DISEASES_URL)
        hpo_disease_genes = retrieve_data(app_settings.HPO_DISEASE_GENES_URL)

        logger.info('Checking HPO Version...')
        version = hpo.meta['data-version'][0]  # NOTE: pronto always returns array
        version_obj, version_created = models.Version.objects.get_or_create(label=version)

        if not version_created:
            logger.info('Version: {0} already synced!'.format(version))
        else:
            logger.info('Syncing new HPO terms with version: {0}'.format(version))

            relations = {}
            for term in hpo:
                identifier = utils.normalize_hpo_id(term.id)
                label = term.name
                description = term.desc
                created_by = term.other.get('created_by')
                created = term.other.get('creation_date')

                alternate_ids = [
                    str(utils.normalize_hpo_id(alt_term))
                    for alt_term in term.other.get('alt_id', [])
                ]

                try:
                    term_obj = models.Term.objects.create(
                        version=version_obj,
                        identifier=identifier,
                        label=label if label else "",
                        description=description if description else "",
                        created_by=created_by[0] if created_by else "",
                        created=created[0] if created else "",
                        alternate_ids=",".join(alternate_ids),
                    )
                except Exception as error:
                    msg = 'HPO: {0} could not be added! Error: {1}'.format(term.id, error)
                    logger.error(msg)
                    raise Exception(msg)

                # Create Synonym objects
                for synonym in term.synonyms:
                    if synonym:
                        synonym_obj, synonym_created = models.Synonym.objects.get_or_create(
                            hpo_term=term_obj,
                            description=synonym.desc if synonym.desc else "",
                            # NOTE: pronto always returns array
                            scope=getattr(choices.SYNONYM_SCOPES, synonym.scope),
                        )

                # Save is_a/can_be relations to add the ManyToManyFields once completed
                relations[term_obj.id] = {'is_a': [], 'can_be': []}
                for key, value in term.relations.items():
                    if str(key) == "Relationship('is_a')":
                        for is_a_term in value:
                            relations[term_obj.id]['is_a'].append(utils.normalize_hpo_id(is_a_term.id))
                    elif str(key) == "Relationship('can_be')":
                        for can_be_term in value:
                            relations[term_obj.id]['can_be'].append(utils.normalize_hpo_id(can_be_term.id))
                    else:
                        err = "Parsing {0} is not implemented!".format(str(key))
                        raise NotImplementedError(err)

                # Create CrossReference objects
                for xref in term.other.get('xref', []):
                    if xref:
                        xref_data = xref.split(':', 1)

                        # NOTE: We want source and id to store. There are 3 records that don't conform
                        if len(xref_data) != 2 or xref_data[0].upper() == 'HTTP':
                            err_msg = 'Format of XRef: {0} for {1} is not supported!'.format(xref, term.id)
                            logger.warning(err_msg)
                            continue

                        xref_obj, xref_created = models.CrossReference.objects.get_or_create(
                            hpo_term=term_obj,
                            # NOTE: pronto always returns array
                            source=getattr(choices.CROSS_REFERENCES, str(xref_data[0].upper())),
                            source_value=xref_data[1],
                        )

            # Create is_a/can_be relationships
            logger.info('Building is_a/can_be relationships...')
            for key, value in relations.items():
                term_obj = models.Term.objects.get(pk=key)

                is_a_objects = models.Term.objects.filter(
                    Q(version=version_obj) &
                    Q(identifier__in=value['is_a'])
                )
                term_obj.is_a.add(*is_a_objects)

                can_be_objects = models.Term.objects.filter(
                    Q(version=version_obj) &
                    Q(identifier__in=value['can_be'])
                )
                term_obj.can_be.add(*can_be_objects)

                term_obj.save()

                if len(is_a_objects) != len(value['is_a']):
                    msg = 'Syncing is_a for Term: {0} failed! {1} vs {2}'.format(
                        key, is_a_objects, value
                    )
                    logger.error(msg)
                    raise ValueError(msg)

                if len(can_be_objects) != len(value['can_be']):
                    msg = 'Syncing can_be for Term: {0} failed! {1} vs {2}'.format(
                        key, can_be_objects, value
                    )
                    logger.error(msg)
                    raise ValueError(msg)

            # Create Disease objects
            logger.info('Syncing diseases from HPO...')
            for line in hpo_diseases:
                disease_record = line.strip().split('\t')

                # NOTE: some OMIM records are screwed up, we skip those
                try:
                    identifier = int(disease_record[1])
                except ValueError:
                    logger.warning('Skipping: {0}'.format(disease_record))
                    continue

                database = getattr(choices.DATABASES, disease_record[0])

                description = disease_record[2]
                qualifier = getattr(choices.QUALIFIERS, disease_record[3], None)
                evidence_code = getattr(choices.EVIDENCE, disease_record[6], None)
                age_of_onset = disease_record[7].strip()
                frequency = disease_record[8].strip()
                create_date = disease_record[12]
                assigned_by = disease_record[13]

                if age_of_onset:
                    try:
                        age_of_onset_obj = models.Term.objects.get(
                            identifier=utils.normalize_hpo_id(age_of_onset)
                        )
                    except ObjectDoesNotExist:
                        age_of_onset_obj = None
                else:
                    age_of_onset_obj = None

                if frequency:
                    try:
                        frequency_obj = models.Term.objects.get(
                            identifier=utils.normalize_hpo_id(frequency)
                        )
                    except ObjectDoesNotExist:
                        frequency_obj = None
                else:
                    frequency_obj = None

                disease_obj, updated = models.Disease.objects.update_or_create(
                    version=version_obj,
                    database=database,
                    identifier=identifier,
                    defaults={
                        'description': description,
                        'qualifier': qualifier,
                        'evidence_code': evidence_code,
                        'age_of_onset': age_of_onset_obj,
                        'frequency': frequency_obj,
                        'create_date': create_date,
                        'assigned_by': assigned_by,
                    }
                )

                hpo_term = models.Term.objects.get(
                    identifier=utils.normalize_hpo_id(disease_record[4])
                )

                disease_obj.hpo_terms.add(hpo_term)
                disease_obj.save()

            # Create Disease Gene objects
            logger.info('Syncing disease genes from HPO')
            for line in hpo_disease_genes:
                if line.startswith('#'):  # NOTE: Skips header
                    continue

                (gene_id, symbol, disease) = line.split('\t')
                (database, identifier) = disease.strip().split(':')

                try:
                    gene_obj = Gene.objects.get(symbol__iexact=symbol)
                except ObjectDoesNotExist:
                    msg = 'Gene: {0} not in the database!'.format(symbol)
                    logger.warning(msg)

                disease_obj, created = models.Disease.objects.get_or_create(
                    version=version_obj,
                    database=getattr(choices.DATABASES, database),
                    identifier=int(identifier),
                    defaults={
                        'create_date': datetime.datetime.now().date()
                    }
                )

                disease_obj.genes.add(gene_obj)
                disease_obj.save()
